#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2017-2020 - The Aura Engine Team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Dealing with configuration data.
"""

import logging
import os
import os.path
import re
import sys
from pathlib import Path

import confuse
import yaml

template = {
    "general": {
        "socket_dir": str,
        "cache_dir": str,
    },
    "log": {
        "directory": str,
        "level": confuse.OneOf(["debug", "info", "warning", "error", "critical"]),
    },
    "monitoring": {
        "heartbeat": {"host": str, "port": int, "frequency": int},
    },
    "api": {
        "steering": {"status": str, "calendar": str},
        "tank": {"session": str, "secret": str, "status": str, "playlist": str},
        "engine": {
            "number": int,
            "status": str,
            "store_playlog": str,
            "store_clock": str,
            "store_health": str,
        },
    },
    "scheduler": {
        "audio": {
            "source_folder": str,
            "source_extension": str,
            "playlist_folder": str,
            "engine_latency_offset": float,
        },
        "fetching_frequency": int,
        "scheduling_window_start": int,
        "scheduling_window_end": int,
        "preload_offset": int,
        "input_stream": {"buffer": float},
        "fade_in_time": float,
        "fade_out_time": float,
    },
}


class AuraConfig:
    """
    Creates config by reading yaml file according to template above.
    """

    _instance = None
    config_file_path = ""
    confuse_config: confuse.Configuration
    config: confuse.Configuration | None = (
        None  # TODO points to a validated config (hopefully later)
    )
    logger: logging.Logger | None = None

    # FIXME: Class properties are deprecated in Python 3.11
    # and will not be supported in Python 3.13
    @classmethod
    @property
    def instance(cls):
        """Create and return singleton instance."""
        if cls._instance is None:
            cls._instance = AuraConfig()
        return cls._instance

    def __init__(self, config_file_path="/etc/aura/engine.yaml"):
        """
        Initialize the configuration, defaults to `/etc/aura/engine.yaml`.

        If this file doesn't exist it uses `./config/engine.yaml` from
        the project directory.

        Args:
            config_file_path(String): The path to the configuration file `engine.yaml`

        """
        self.logger = logging.getLogger("engine")
        config_file = Path(config_file_path)
        project_root = Path(__file__).parent.parent.parent.parent.absolute()

        if not config_file.is_file():
            config_file_path = f"{project_root}/config/engine.yaml"

        self.config_file_path = config_file_path
        print(f"Using configuration at: {config_file_path}")

        envar_matcher = re.compile(r"\$\{([^}^{]+)\}")

        def envar_constructor(loader, node):
            value = os.path.expandvars(node.value)
            # workaround not to parse numerics as strings
            try:
                value = int(value)
            except ValueError:
                pass
            try:
                value = float(value)
            except ValueError:
                pass
            return value

        envar_loader = yaml.SafeLoader
        envar_loader.add_implicit_resolver("!envar", envar_matcher, None)
        envar_loader.add_constructor("!envar", envar_constructor)

        self.confuse_config = confuse.Configuration("engine", loader=envar_loader)
        self.confuse_config.set_file(config_file_path)
        self.load_config()

        # custom overrides and defaults
        self.confuse_config["install_dir"].set(os.path.realpath(project_root))
        self.confuse_config["config_dir"].set(os.path.dirname(config_file_path))

        AuraConfig.instance = self

    def init_version(self, version: dict):
        """
        Read and set the component version from VERSION file in project root.
        """
        self.confuse_config["version_control"].set(version.get("control"))
        self.confuse_config["version_core"].set(version.get("core"))
        self.confuse_config["version_liquidsoap"].set(version.get("liquidsoap"))

    def load_config(self):
        """
        Set config defaults and load settings from file.
        """
        if not os.path.isfile(self.config_file_path):
            self.logger.critical(self.config_file_path + " not found  :(")
            sys.exit(1)

        self.config = self.confuse_config.get(template)
