#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2017-2020 - The Aura Engine Team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
The Engine.
"""

from __future__ import annotations

import json
import logging
import time
from contextlib import suppress
from enum import Enum

import confuse
import tomli

import aura_engine.events as events
import aura_engine.scheduling.scheduler as scheduler  # noqa
import aura_engine.scheduling.timetable as timetable  # noqa
from aura_engine.base.api import LiquidsoapUtil as LU
from aura_engine.base.config import AuraConfig
from aura_engine.base.lang import DotDict
from aura_engine.base.utils import SimpleUtil as SU
from aura_engine.core.channels import ChannelType, QueueChannel
from aura_engine.core.client import CoreConnectionError, PlayoutClient
from aura_engine.core.mixer import Mixer
from aura_engine.resources import (
    ResourceClass,
    ResourceMapping,
    ResourceType,
    ResourceUtil,
)
from aura_engine.scheduling.domain import PlaylistItem


class Engine:
    """
    The Engine.
    """

    instance: Engine = None
    logger: logging.Logger
    config: confuse.Configuration
    engine_time_offset = 0.0
    scheduler: scheduler.AuraScheduler = None  # noqa: F811
    event_dispatcher: events.EngineEventDispatcher = None
    playout: PlayoutClient | None = None
    playout_state = None

    def __init__(self):
        """
        Initialize the engine.
        """
        if Engine.instance:
            raise Exception("Engine is already running!")
        Engine.instance = self
        self.logger = logging.getLogger("engine")
        self.config = AuraConfig.instance.config
        Engine.engine_time_offset = self.config.scheduler.audio.engine_latency_offset

    def start(self):
        """
        Start the engine.

        Called when the connection to the sound-system implementation has been established.
        """
        from aura_engine.plugins.clock import ClockInfoHandler
        from aura_engine.plugins.monitor import AuraMonitor

        self.event_dispatcher = events.EngineEventDispatcher(self)

        # Subscribe event handlers to event dispatcher events
        binding = self.event_dispatcher.attach(AuraMonitor)
        binding.subscribe("on_boot")
        binding.subscribe("on_sick")
        binding.subscribe("on_resurrect")
        binding = self.event_dispatcher.attach(ClockInfoHandler)
        binding.subscribe("on_play")
        binding.subscribe("on_fallback_active")

        from aura_engine.scheduling.scheduler import AuraScheduler

        scheduler = AuraScheduler(self)
        self.event_dispatcher.initialize(scheduler)

        while not self.is_connected():
            self.logger.info(SU.yellow("Waiting for Liquidsoap to be running ..."))
            time.sleep(2)

        self.logger.info(SU.green("Engine Core ------[ connected ]-------- Liquidsoap"))
        self.event_dispatcher.on_boot()

        self.logger.info(EngineSplash.splash_screen())
        self.event_dispatcher.on_ready()

    #
    #   Basic Methods
    #

    def is_connected(self):
        """
        Check if there's a valid connection to Liquidsoap.
        """
        has_connection = False
        try:
            if not self.playout:
                self.playout = PlayoutClient.get_instance(self.event_dispatcher)
            self.playout.get_uptime()
            self.logger.info(SU.green("Initialize Player..."))
            self.player = Player(self.playout, self.event_dispatcher)
            has_connection = True
        except CoreConnectionError:
            self.logger.debug("Liquidsoap is not running so far")
        except Exception as e:
            self.logger.error(f"Cannot check if Liquidsoap is running. \nReason: {e}", e)

        return has_connection

    def update_playout_state(self):
        """
        Retrieve the state of all inputs and outputs.
        """
        state = self.playout.get_status()
        state = DotDict(LU.json_to_dict(state))

        def dispatch_fallback_event():
            timetable: timetable.TimetableService = self.scheduler.get_timetable()  # noqa: F811
            timeslot = timetable.get_current_timeslot()
            self.event_dispatcher.on_fallback_active(timeslot)

        # Initialize state
        if not self.playout_state:
            if state.is_fallback:
                self.logger.info(SU.yellow("Set initial playout state to FALLBACK"))
                dispatch_fallback_event()
            else:
                self.logger.info(SU.green("Set initial playout state"))
        elif state and self.playout_state.is_fallback != state.is_fallback:
            if state.is_fallback:
                self.logger.info(SU.yellow("Playout turned into FALLBACK state"))
            else:
                self.logger.info(SU.green("Playout turned back into normal state"))

        self.playout_state = state
        return self.playout_state

    def get_playout_state(self):
        """
        Retrieve the state of Engine Core.
        """
        return self.playout_state

    def init_version(self) -> dict:
        """
        Get the versions of Engine components and store in configuration.

        Returns:
            dict: {
                "control": Engine Control version
                "core": Engine Core version
                "liquidsoap": Liquidsoap version
            }

        """
        with open("pyproject.toml", mode="rb") as config:
            toml_file = tomli.load(config)
        ctrl_version = toml_file["tool"]["poetry"]["version"]
        versions = self.playout.get_version()
        versions = DotDict(json.loads(versions))
        versions.control = ctrl_version
        AuraConfig.instance.init_version(versions)

    @staticmethod
    def engine_time():
        """
        Get the engine perspective on time.

        Liquidsoap is slow in executing commands, therefore it's needed to schedule
        actions by (n) seconds in advance, as defined in the configuration file by
        the property `engine_latency_offset`. it's important to note that this method
        requires the class variable `EngineUtil.engine_time_offset` to be set on
        Engine initialization.

        Returns:
            (Integer):  the Unix epoch timestamp including the offset

        """
        return SU.timestamp() + Engine.engine_time_offset

    @staticmethod
    def get_instance() -> Engine:
        """
        Return the one and only engine.

        Create an instance, if it doesn't exist yet.

        Returns:
            (Engine): The engine instance.
        """
        if not Engine.instance:
            Engine.instance = Engine()
        return Engine.instance

    def terminate(self):
        """
        Terminate the engine and all related processes.
        """
        if self.scheduler:
            self.scheduler.terminate()
        # TODO terminate core connection


#
#   PLAYER
#


class Player:
    """
    Engine Player.
    """

    class TransitionType(str, Enum):
        """
        Types for instant and fade transitions.
        """

        INSTANT = "instant"
        FADE = "fade"

    logger: logging.Logger
    config: confuse.Configuration
    channels = None
    resource_map: ResourceMapping
    event_dispatcher: events.EngineEventDispatcher
    mixer: Mixer

    def __init__(self, playout: PlayoutClient, event_dispatcher: events.EngineEventDispatcher):
        """
        Initialize the player.

        Args:
            playout (PlayoutClient): Client for connecting to Engine Core (Liquidsoap)
            event_dispatcher (EventDispatcher): Dispatcher for issuing events

        """
        self.config = AuraConfig.instance.config
        self.logger = logging.getLogger("engine")
        self.event_dispatcher = event_dispatcher
        self.resource_map = ResourceMapping()
        # self.mixer = playout.get_mixer()
        self.mixer = Mixer("mixer", playout)

    def preload(self, item: PlaylistItem):
        """
        Pre-load the item. This is required before the actual `play(..)` can happen.

        Be aware when using this method in order to queue a very short item this may
        result in situations with incorrect timing. In this case bundle multiple short
        items as one queue using `preload_playlist(self, items)`.

        It's important to note, that his method is blocking until loading has finished. If this
        method is called asynchronously, the progress on the preloading state can be looked up in
        the `item.play` state object.

        Args:
            item (PlaylistItem): An array holding filesystem items

        """
        is_queue = item.get_content_type() in ResourceClass.FILE.types
        metadata = ResourceUtil.generate_track_metadata(item, not is_queue)
        self.logger.debug(SU.pink(f"Loading item '{item}'"))
        uri = None
        is_ready = False

        # LINE
        if item.get_content_type() in ResourceClass.LIVE.types:
            channel_name = self.resource_map.live_channel_for_resource(item.source)
            chosen_channel = self.mixer.get_channel(channel_name)
        else:
            channel_type = self.resource_map.type_for_resource(item.get_content_type())
            chosen_channel = self.mixer.get_free_channel(channel_type)

            # QUEUE
            if is_queue:
                uri = ResourceUtil.source_to_filepath(item.source)
            # STREAM
            elif item.get_content_type() in ResourceClass.STREAM.types:
                uri = item.source

        if not chosen_channel:
            self.logger.critical(SU.red("No channel for '{item.source}' source found"))
        else:
            msg = f"Assign channel {chosen_channel} to item"
            item.play.set_loading(chosen_channel)
            self.logger.info(SU.pink(msg))

        is_ready = item.play.channel.load(uri, metadata=metadata)

        if is_ready:
            item.play.set_ready()
        self.event_dispatcher.on_queue([item])

    def preload_group(self, items: list):
        """
        Preload multiple filesystem/queue items at once.

        This call is required before the actual `play(..)` can happen. Due to their nature,
        non-filesystem items cannot be queued using this method. In this case use
        `preload(self, item)` instead. This method also allows queuing of very short files, such
        as jingles.

        It's important to note, that his method is blocking until loading has finished. If this
        method is called asynchronously, the progress on the preloading state can be looked up in
        `item.state`.

        Args:
            items ([PlaylistItem]): list of playlist items.

        """
        free_channel = self.mixer.get_free_channel(ChannelType.QUEUE)

        # Validate item type
        for item in items:
            if item.get_content_type() != ResourceType.FILE:
                raise InvalidChannelException

        # Determine channel & queue items
        item: PlaylistItem
        for item in items:
            item.play.set_loading(free_channel)
            self.logger.debug(SU.pink(f"Loading item '{item}'"))

            # Choose and save the input channel
            metadata = ResourceUtil.generate_track_metadata(item)
            file_path = ResourceUtil.source_to_filepath(item.source)
            if item.play.channel.load(file_path, metadata):
                item.play.set_ready()

        self.event_dispatcher.on_queue(items)

    def play(self, item: PlaylistItem, transition: TransitionType):
        """
        Play a new `PlaylistItem`.

        In case of a new timeslot (or some intended, immediate transition), a prepared channel is
        selected and transitions between old and new channel is performed.

        This method expects that the item is pre-loaded using `preload(..)` or
        `preload_group(self, items)` before being played. In case the pre-loading has happened
        for a group of items, only the first item of the group needs to be passed.

        Args:
            item (PlaylistItem): The audio source to be played
            transition (Player.TransitionType): The type of transition to use e.g. fade-in or
                instant volume level.
            queue (Boolean): If `True` the item is queued if the `ChannelType` does allow so;
                otherwise a new channel of the same type is activated

        TODO The command `item.play.set_playing()` sets the playlist item's status to _playing_
        and assigns the current play start time. This is required for resource like line-in.
        Other source like audio files have their start marker natively assigned by Liquidsoap.
        These timestamp represent the actual real start time and should somehow be retrieved
        from Liquidsoap, in order to also have the _real deal_ in Engine.
        """
        with suppress(CoreConnectionError):
            # Stop any active channel
            self.stop(Player.TransitionType.FADE)

            # Start the new channel
            item.play.set_playing()
            instant = not (transition == Player.TransitionType.FADE)
            item.play.channel.fade_in(item.volume, instant=instant)

            # Dispatch event
            self.event_dispatcher.on_play(item)

    def stop(self, transition: TransitionType):
        """
        Stop the currently active channel either instantly or by fading out.

        Args:
            transition (Player.TransitionType): The type of transition to use e.g. fade-out
        """
        with suppress(CoreConnectionError):
            channel = self.mixer.get_active_channel()
            if not channel:
                self.logger.info(SU.pink("Nothing to stop, no active channel"))
                return

            instant = not (transition == Player.TransitionType.FADE)
            channel.fade_out(instant=instant)
            self.event_dispatcher.on_stop(channel)

    def roll(self, channel: QueueChannel, seconds: int) -> bool:
        """
        Pre-roll the player of the given `ChannelType.QUEUE` channel by (n) seconds.

        Args:
            channel (ChannelName): The channel to push the file to
            seconds (Float): The seconds to roll

        Returns:
            (bool): True after successful rolling
        """
        return channel.roll(seconds)


class InvalidChannelException(Exception):
    """
    Exception thrown when the given channel is invalid.
    """

    pass


class EngineSplash:
    """Print the splash and version information on boot."""

    @staticmethod
    def splash_screen():
        """
        Print the engine logo and version info.
        """
        version = AuraConfig.instance.confuse_config["version_control"].get()
        core_version = AuraConfig.instance.confuse_config["version_core"].get()
        liq_version = AuraConfig.instance.confuse_config["version_liquidsoap"].get()

        return f"""\n
             █████╗ ██╗   ██╗██████╗  █████╗     ███████╗███╗   ██╗ ██████╗ ██╗███╗   ██╗███████╗
            ██╔══██╗██║   ██║██╔══██╗██╔══██╗    ██╔════╝████╗  ██║██╔════╝ ██║████╗  ██║██╔════╝
            ███████║██║   ██║██████╔╝███████║    █████╗  ██╔██╗ ██║██║  ███╗██║██╔██╗ ██║█████╗
            ██╔══██║██║   ██║██╔══██╗██╔══██║    ██╔══╝  ██║╚██╗██║██║   ██║██║██║╚██╗██║██╔══╝
            ██║  ██║╚██████╔╝██║  ██║██║  ██║    ███████╗██║ ╚████║╚██████╔╝██║██║ ╚████║███████╗
            ╚═╝  ╚═╝ ╚═════╝ ╚═╝  ╚═╝╚═╝  ╚═╝    ╚══════╝╚═╝  ╚═══╝ ╚═════╝ ╚═╝╚═╝  ╚═══╝╚══════╝
            control v{version}, core v{core_version}, liquidsoap v{liq_version} - Ready to play!
        \n"""
