#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2017-2020 - The Aura Engine Team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Utilities and mappings for media resources.
"""

from __future__ import annotations

import datetime
from enum import Enum

from aura_engine.base.config import AuraConfig
from aura_engine.base.utils import SimpleUtil as SU
from aura_engine.core.channels import ChannelName, ChannelType
from aura_engine.scheduling import domain


class ResourceType(Enum):
    """
    Media content types.
    """

    FILE = "file:"
    STREAM_HTTP = "http"
    LINE = "line:"
    M3U = "m3u:"
    POOL = "pool:"


class ResourceMapping:
    """
    Wires source types with channel-types.
    """

    resource_mapping = None

    def __init__(self):
        """
        Initialize the resource mapping.
        """

        self.resource_mapping = {
            ResourceType.FILE: ChannelType.QUEUE,
            ResourceType.STREAM_HTTP: ChannelType.HTTP,
            ResourceType.LINE: ChannelType.LIVE,
            ResourceType.M3U: ChannelType.QUEUE,
            ResourceType.POOL: ChannelType.QUEUE,
        }

    def type_for_resource(self, resource_type: ResourceType) -> ChannelType:
        """
        Retrieve a `ChannelType` for the given `ResourceType`.

        Only default mappings can be evaluated. Custom variations
        like fallback channels are not respected.
        """
        return self.resource_mapping.get(resource_type)

    def live_channel_for_resource(self, channel_uri: str) -> ChannelName:
        """
        Return the channel enum for a given live channel string from Tank.

        Channel URIs from Tank are typically in the format `line://1`, while
        channel names in Liquidsoap have names in the format `aura_engine_line_in_1`.

        If no valid channel URI is provided, `None` is returned.

        Args:
            channel (str): Channel URI as provided from Tank.

        Returns:
            (ChannelName): Liquidsoap name of the channel.
        """
        if not channel_uri:
            return None
        channel_name = "aura_engine_line_in_" + channel_uri.split("line://")[1]

        for cn in ChannelName:
            if cn.value == channel_name:
                return cn

        return None


class ResourceClass(Enum):
    """
    Media content classes.
    """

    FILE = {"id": "fs", "numeric": 0, "types": [ResourceType.FILE]}
    STREAM = {
        "id": "http",
        "numeric": 1,
        "types": [ResourceType.STREAM_HTTP],
    }
    LIVE = {"id": "live", "numeric": 2, "types": [ResourceType.LINE]}
    PLAYLIST = {
        "id": "playlist",
        "numeric": 3,
        "types": [ResourceType.M3U, ResourceType.POOL],
    }

    @property
    def types(self):
        """Retrieve allowed types of the resource class."""
        return self.value["types"]

    @property
    def numeric(self):
        """Retrieve a numeric representation of the resource class."""
        return self.value["numeric"]

    def __str__(self):
        return str(self.value["id"])


class ResourceUtil(Enum):
    """
    Utilities for different resource types.
    """

    @staticmethod
    def get_content_type(uri: str) -> ResourceType:
        """
        Return the content type identified by the passed URI.

        Args:
            uri (str): The URI of the source.

        Returns:
            (ResourceType): The identified resource type.

        """
        if not uri:
            return None
        if uri.startswith(ResourceType.STREAM_HTTP.value):
            return ResourceType.STREAM_HTTP
        if uri.startswith(ResourceType.POOL.value):
            return ResourceType.POOL
        if uri.startswith(ResourceType.M3U.value):
            return ResourceType.M3U
        if uri.startswith(ResourceType.FILE.value):
            return ResourceType.FILE
        if uri.startswith(ResourceType.LINE.value):
            return ResourceType.LINE

    @staticmethod
    def get_content_class(content_type: ResourceType) -> ResourceClass:
        """
        Return the content class identified by the passed type.

        Args:
            content_type (ContentType): The resource type.

        Returns:
            (ResourceClass): The class this type falls in.

        """
        if content_type in ResourceClass.FILE.types:
            return ResourceClass.FILE
        if content_type in ResourceClass.STREAM.types:
            return ResourceClass.STREAM
        if content_type in ResourceClass.LIVE.types:
            return ResourceClass.LIVE
        if content_type in ResourceClass.PLAYLIST.types:
            return ResourceClass.PLAYLIST

    @staticmethod
    def generate_m3u_file(target_file_path: str, items: list):
        """
        Write a M3U file based on the given playlist object.

        Args:
            target_file_path (str): The M3U playlist to write.
            items ([PlaylistItem]): Items of the playlist.
            config (AuraConfig): The config file wrapper.
        """
        file = open(target_file_path, "w")
        fb = ["#EXTM3U\n"]

        for item in items:
            if ResourceUtil.get_content_type(item.source) == ResourceType.FILE:
                path = ResourceUtil.source_to_filepath(item.source)
                fb.append(
                    f"#EXTINF:{item.duration},{item.metadata.artist} - {item.metadata.title}"
                )
                fb.append(path + "\n")

        file.writelines(fb)
        file.close()

    @staticmethod
    def to_abs_path(path):
        """
        Transform any given (relative) path to an absolute path.

        Starting at the project root.
        """
        if path.startswith("/"):
            return path
        else:
            ac = AuraConfig.instance
            return ac.confuse_config["install_dir"].get() + "/" + path

    @staticmethod
    def audio_store_path() -> str:
        """
        Get audio store location.

        Returns:
            (str): Absolute path to audio store folder.
        """
        ac = AuraConfig.instance
        return ResourceUtil.to_abs_path(ac.config.scheduler.audio.source_folder)

    @staticmethod
    def playlist_folder_path() -> str:
        """
        Get playlist folder location.

        Returns:
            (str): Return the absolute path to the playlist folder.
        """
        ac = AuraConfig.instance
        return ResourceUtil.to_abs_path(ac.config.scheduler.audio.playlist_folder)

    @staticmethod
    def source_to_filepath(source_uri: str) -> str:
        """
        Create path from URI and extension.

        Convert a file-system URI starting with "file://" to an actual, absolute path to the file,
        appending the extension as provided in "source_extension".

        If the path starts with an "/", it indicates that it is already an absolute path including
        a valid extension.

        Args:
            source_uri (str): The URI of the file

        Returns:
            (str): Absolute file path.

        """
        path = source_uri[7:]
        if path.startswith("/"):
            return path
        else:
            ac = AuraConfig.instance
            base_dir = ResourceUtil.audio_store_path()
            extension = ac.config.scheduler.audio.source_extension
            return base_dir + "/" + path + extension

    @staticmethod
    def get_items_string(items: list) -> str:
        """
        Return a list of playlist items as String for logging purposes.

        Args:
            items ([PlaylistItem]): List of playlist items.

        Returns:
            (str): Stringified version of the items.
        """
        s = ""
        if isinstance(items, list):
            for item in items:
                s += str(item)
                if item != items[-1]:
                    s += ", "
        else:
            s = str(items)
        return s

    @staticmethod
    def generate_track_metadata(
        item: domain.PlaylistItem,
        assign_track_start: bool = False,
    ) -> dict:
        """
        Generate Liquidsoap track metadata based on an playlist item.

        Args:
            item (PlaylistItem): The playlist item.
            assign_track_start (bool): If true, the track start time metadata is passed through
                to Liquidsoap. This is recommended for audio sourced where Liquidsoap does not
                generate its own start time (e.g line from live studio).

        Returns:
            ({}): dictionary with structured metadata.
        """
        content_type = ResourceUtil.get_content_type(item.source)
        content_class = ResourceUtil.get_content_class(content_type)
        playlist: domain.Playlist = item.get_playlist()
        timeslot: domain.Timeslot = playlist.get_timeslot()

        try:
            annotations = {
                "show_name": str(timeslot.get_show().name),
                "show_id": int(timeslot.get_show().id),
                "timeslot_id": timeslot.id,
                "playlist_id": int(playlist.id),
                "playlist_item": str(float(item.get_position())),
                "track_type": int(content_class.numeric),
                "track_start": "",
                "track_duration": -1,
                "track_title": "",
                "track_album": "",
                "track_artist": "",
            }
        except ValueError as ve:
            raise ve

        if assign_track_start:
            # Convert to current Liquidsoap date/time format
            track_start: datetime.datetime = SU.timestamp_to_datetime(item.get_start())
            annotations["track_start"] = track_start.strftime("%Y/%m/%d %H:%M:%S")
        if item.duration and item.duration > 0:
            annotations["track_duration"] = item.duration
        if item.metadata:
            annotations["track_title"] = str(item.metadata.title)
            annotations["track_album"] = str(item.metadata.album)
            annotations["track_artist"] = str(item.metadata.artist)

        return annotations
