#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2017-now() - The Aura Engine Team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import logging
from enum import StrEnum, auto
from time import sleep
from urllib.parse import urlparse

from aura_engine.base.config import AuraConfig
from aura_engine.base.lang import synchronized
from aura_engine.base.utils import SimpleUtil as SU
from aura_engine.core.channels import ChannelName, PlayoutStatusResponse
from aura_engine.core.client import CoreClient
from aura_engine.core.mixer import Mixer
from aura_engine.events import EngineEventDispatcher


# enums reflecting namespace, action and reponses
class Namespace(StrEnum):
    MIXER = auto()
    CHANNEL = auto()

    @classmethod
    def _missing_(cls, value):
        # map 'mixer' to .MIXER and values of ChannelName to .CHANNEL
        if isinstance(value, Namespace):
            return cls(value)
        elif value in ChannelName._value2member_map_.values():
            return cls.CHANNEL
        else:
            raise TypeError


class MixerAction(StrEnum):
    INPUTS = auto()
    OUTPUTS = auto()
    STATUS = auto()
    VOLUME = auto()
    SELECT = auto()
    ACTIVATE = auto()


class ChannelAction(StrEnum):
    PUSH = auto()
    LOAD = auto()
    URL = auto()
    STATUS = auto()
    START = auto()
    STOP = auto()
    CLEAR = auto()
    ROLL = auto()
    SET_TRACK_METADATA = auto()


class CoreClientMock(CoreClient):
    """
    Suppress unwanted behavior by subclassing and overriding necessary functions.
    """

    instance = None
    connection = None
    event_dispatcher = None
    conn = None
    stream_url = None
    stream_status = PlayoutStatusResponse.STREAM_STATUS_CONNECTED
    resource_id = -1
    volumes = [0, 0, 0, 0, 0, 0]
    selections = [0, 0, 0, 0, 0, 0]

    def __init__(self, event_dispatcher: EngineEventDispatcher):
        """
        Initialize the client.
        """
        self.config = AuraConfig.instance.config
        self.event_dispatcher = event_dispatcher

    @staticmethod
    def get_instance(event_dispatcher: EngineEventDispatcher):
        """
        Get an instance of the client singleton.
        """
        if not CoreClientMock.instance:
            CoreClientMock.instance = CoreClientMock(event_dispatcher)
        return CoreClientMock.instance

    def _validate_url(self, url: str) -> bool:
        """
        Check url validity.
        """
        parse = urlparse(url)
        return all([parse.scheme, parse.netloc])

    def _get_resource_id(self) -> int:
        """
        Increment and return index starting with 0.
        """
        self.resource_id += 1
        return self.resource_id

    @synchronized
    def connect(self):
        sleep(0.5)
        pass

    @synchronized
    def disconnect(self):
        sleep(0.5)
        pass

    @synchronized
    def exec(self, namespace: str, action: str, args: str = "") -> str:
        log = print
        # log(f"Core mock request 'ns: {namespace}', action: '{action}', args: '{args}'")

        response: str = None

        ns = Namespace(namespace)
        if ns is None:
            raise TypeError
        match ns:
            case Namespace.MIXER:
                act = MixerAction(action)
                if act is None:
                    raise TypeError
                match act:
                    case MixerAction.INPUTS:
                        response = f"{ChannelName.QUEUE_A} {ChannelName.QUEUE_B} {ChannelName.HTTP_A} {ChannelName.HTTP_B} {ChannelName.LIVE_0} {ChannelName.LIVE_1}"
                    case MixerAction.OUTPUTS:
                        response = '["aura_engine_line_out_0", "out_http_0"]'  # no abstraction yet
                    case MixerAction.STATUS:
                        chn = int(args)
                        vol = self.volumes[chn]
                        response = (
                            f"ready=true selected=true single=false volume={vol}% remaining=0"
                        )
                    case MixerAction.VOLUME:
                        argv = args.split(" ")
                        chn = int(argv[0])
                        vol = int(float(argv[1]) * 100.0)
                        self.volumes[chn] = vol
                        response = f"volume={vol}% remaining=0"
                    case MixerAction.SELECT:
                        argv = args.split(" ")
                        chn = int(argv[0])
                        vol = self.volumes[chn]
                        sel = 1 if argv[1] == "true" else 0
                        self.selections[chn] = sel
                        response = f"volume={vol}% remaining=0 selected={sel}"
                    case MixerAction.ACTIVATE:
                        response = PlayoutStatusResponse.SUCCESS
            case Namespace.CHANNEL:
                act = ChannelAction(action)
                if act is None:
                    raise TypeError
                match act:
                    case ChannelAction.PUSH:
                        sleep(5)
                        resid = self._get_resource_id()
                        response = str(resid)
                    case ChannelAction.LOAD:
                        sleep(3)
                        resid = self._get_resource_id()
                        response = str(resid)
                    case ChannelAction.URL:
                        sleep(0.5)
                        self.stream_url = args
                        response = PlayoutStatusResponse.SUCCESS
                    case ChannelAction.STATUS:
                        if self._validate_url(self.stream_url):
                            response = self.stream_status
                            if response == PlayoutStatusResponse.STREAM_STATUS_CONNECTED:
                                response += " " + self.stream_url
                        else:
                            # simulate polling state by supplying invalid url
                            response = PlayoutStatusResponse.STREAM_STATUS_POLLING
                    case ChannelAction.START:
                        sleep(1)
                        response = PlayoutStatusResponse.SUCCESS
                    case ChannelAction.STOP:
                        response = PlayoutStatusResponse.SUCCESS
                    case ChannelAction.CLEAR:
                        response = PlayoutStatusResponse.SUCCESS
                    case ChannelAction.ROLL:
                        response = PlayoutStatusResponse.SUCCESS
                    case ChannelAction.SET_TRACK_METADATA:
                        response = PlayoutStatusResponse.SUCCESS

        if response is not None:
            # log(f"Core mock reponse '{response}'")
            return response
        raise Exception(f"Unhandled namespace: '{namespace}', action: '{action}', args: '{args}'")


class PlayoutClientMock(CoreClientMock):
    """
    Mock PlayoutClient by subclssing and skipping unwanted.
    """

    mixer: Mixer = None

    def __init__(self, event_dispatcher: EngineEventDispatcher):
        super().__init__(event_dispatcher)
        self.mixer = Mixer("mixer", self)

    @staticmethod
    def get_instance(event_dispatcher: EngineEventDispatcher) -> CoreClientMock:
        if not PlayoutClientMock.instance:
            PlayoutClientMock.instance = PlayoutClientMock(event_dispatcher)
        return PlayoutClientMock.instance

    def get_mixer(self):
        return self.mixer
