#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2017-now() - The Aura Engine Team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import unittest

from aura_engine.base.config import AuraConfig
from aura_engine.base.logger import AuraLogger
from aura_engine.core.channels import (
    ChannelFactory,
    ChannelName,
    ChannelType,
    LineChannel,
    PlayoutStatusResponse,
    QueueChannel,
    StreamChannel,
)
from aura_engine.core.mixer import Mixer
from aura_engine.events import EngineEventDispatcher
from tests.core_client_mock import PlayoutClientMock


class TestChannelFactory(unittest.TestCase):
    """
    Testing the ChannelFactory.
    """

    mixer: Mixer = None
    factory: ChannelFactory = None

    def setUp(self):
        config = AuraConfig.instance.config
        AuraLogger(config)

        dispatcher = EngineEventDispatcher(engine=None)
        client = PlayoutClientMock(event_dispatcher=dispatcher)
        self.mixer = client.mixer
        self.factory = ChannelFactory(self.mixer)

    def create_channels(self):
        queue = self.factory.create_channel(0, ChannelName.QUEUE_A, self.mixer)
        stream = self.factory.create_channel(0, ChannelName.HTTP_A, self.mixer)
        live = self.factory.create_channel(0, ChannelName.LIVE_0, self.mixer)
        return (queue, stream, live)

    def test_create_channel(self):
        print(self._testMethodName)
        queue, stream, live = self.create_channels()
        self.assertEqual(type(queue), QueueChannel)
        self.assertEqual(type(stream), StreamChannel)
        self.assertEqual(type(live), LineChannel)

    def test_get_type(self):
        print(self._testMethodName)
        queue, stream, live = self.create_channels()
        self.assertEqual(queue.get_type(), ChannelType.QUEUE)
        self.assertEqual(stream.get_type(), ChannelType.HTTP)
        self.assertEqual(live.get_type(), ChannelType.LIVE)
        self.assertEqual(queue.get_type().numeric, ChannelType.QUEUE.numeric)
        self.assertEqual(stream.get_type().numeric, ChannelType.HTTP.numeric)
        self.assertEqual(live.get_type().numeric, ChannelType.LIVE.numeric)

    def test_get_index(self):
        print(self._testMethodName)
        queue = self.factory.create_channel(0, ChannelName.QUEUE_A, self.mixer)
        stream = self.factory.create_channel(1, ChannelName.HTTP_A, self.mixer)
        live = self.factory.create_channel(2, ChannelName.LIVE_0, self.mixer)
        self.assertEqual(queue.get_index(), 0)
        self.assertEqual(stream.get_index(), 1)
        self.assertEqual(live.get_index(), 2)

    def test_load(self):
        print(self._testMethodName)
        queue, stream, live = self.create_channels()
        queue.load("./file.flac")
        stream.load("http://stream.local")
        live.load()

    def test_load_metadata(self):
        print(self._testMethodName)
        queue, stream, live = self.create_channels()
        metadata = {"title": "A Title", "artist": "An Artist", "album": "An Album"}
        queue.load("./file.flac", metadata=metadata)
        stream.load("http://stream.local", metadata=metadata)
        live.load(metadata=metadata)

    def test_fade_in_out(self):
        print(self._testMethodName)
        queue, stream, live = self.create_channels()

        queue.fade_out()
        stream.fade_out()
        live.fade_out()

        queue.fade_in(volume=100, instant=True)
        stream.fade_in(volume=100)
        queue.fade_out()

        live.fade_in(volume=100)
        stream.fade_out()
        live.fade_out(instant=True)

    def test_roll(self):
        print(self._testMethodName)
        queue = self.factory.create_channel(0, ChannelName.QUEUE_A, self.mixer)
        res = queue.roll(300)
        self.assertTrue(res)


class TestPlayoutStatusResponse(unittest.TestCase):
    """
    Testing the PlayoutStatusResponse.
    """

    def test_enum(self):
        print(self._testMethodName)

        self.assertIs(PlayoutStatusResponse("OK"), PlayoutStatusResponse.SUCCESS)
        self.assertIs(PlayoutStatusResponse("Done"), PlayoutStatusResponse.SUCCESS)
        self.assertIs(PlayoutStatusResponse("Done!"), PlayoutStatusResponse.SUCCESS)
        self.assertIs(PlayoutStatusResponse("Donee!"), PlayoutStatusResponse.SUCCESS)
        self.assertIs(
            PlayoutStatusResponse("polling"), PlayoutStatusResponse.STREAM_STATUS_POLLING
        )
        self.assertIs(
            PlayoutStatusResponse("stopped"), PlayoutStatusResponse.STREAM_STATUS_STOPPED
        )
        self.assertIs(
            PlayoutStatusResponse("connected <additional data>"),
            PlayoutStatusResponse.STREAM_STATUS_CONNECTED,
        )
        self.assertIs(PlayoutStatusResponse("mkay"), PlayoutStatusResponse.INVALID)
        self.assertIs(PlayoutStatusResponse(""), PlayoutStatusResponse.INVALID)


if __name__ == "__main__":
    unittest.main()
