#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2017-now() - The Aura Engine Team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import unittest

from aura_engine.base.config import AuraConfig
from aura_engine.base.logger import AuraLogger
from aura_engine.core.channels import (
    ChannelName,
    ChannelType,
    LineChannel,
    QueueChannel,
    StreamChannel,
)
from aura_engine.core.mixer import Mixer
from aura_engine.events import EngineEventDispatcher
from tests.core_client_mock import PlayoutClientMock


class TestMixer(unittest.TestCase):
    """
    Testing the Mixer.
    Some mixer functions are already covered by TestChannelFactory.
    See test_core_channels.py for details.
    """

    mixer: Mixer = None

    def setUp(self):
        config = AuraConfig.instance.config
        AuraLogger(config)

        dispatcher = EngineEventDispatcher(engine=None)
        client = PlayoutClientMock(event_dispatcher=dispatcher)
        self.mixer = client.mixer

    def test_get_inputs(self):
        print(self._testMethodName)
        inputs = self.mixer.get_inputs()
        input_state = {
            "ready": True,
            "selected": True,
            "single": True,
            "volume": 0,
            "remaining": 0.0,
        }
        self.assertEqual(len(inputs), 6)
        self.assertEqual(inputs["in_queue_0"], input_state)
        self.assertEqual(inputs["in_queue_1"], input_state)
        self.assertEqual(inputs["in_stream_0"], input_state)
        self.assertEqual(inputs["in_stream_1"], input_state)
        self.assertEqual(inputs["aura_engine_line_in_0"], input_state)
        self.assertEqual(inputs["aura_engine_line_in_1"], input_state)

    def test_get_outputs(self):
        print(self._testMethodName)
        outputs = self.mixer.get_outputs()
        self.assertEqual(len(outputs), 2)
        self.assertEqual(outputs[0], "aura_engine_line_out_0")
        self.assertEqual(outputs[1], "out_http_0")

    def test_get_free_channel(self):
        print(self._testMethodName)
        free = self.mixer.get_free_channel(channel_type=ChannelType.LIVE)
        self.assertTrue(isinstance(free, LineChannel))

    def test_get_set_active_channel(self):
        active = self.mixer.get_active_channel()
        self.assertIsNone(active)

        queue_channel = self.mixer.get_channel(channel_name=ChannelName.QUEUE_A)
        self.mixer.set_active_channel(channel=queue_channel)
        active = self.mixer.get_active_channel()
        self.assertTrue(isinstance(active, QueueChannel))

        stream_channel = self.mixer.get_channel(channel_name=ChannelName.HTTP_A)
        self.mixer.set_active_channel(channel=stream_channel)
        active = self.mixer.get_active_channel()
        self.assertTrue(isinstance(active, StreamChannel))

        line_channel = self.mixer.get_channel(channel_name=ChannelName.LIVE_1)
        self.mixer.set_active_channel(channel=line_channel)
        active = self.mixer.get_active_channel()
        self.assertTrue(isinstance(active, LineChannel))


if __name__ == "__main__":
    unittest.main()
