#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2017-now() - The Aura Engine Team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
import unittest

import validators

from aura_engine.base.config import AuraConfig


class TestConfig(unittest.TestCase):
    """
    Testing the Configuration.
    """

    config = None

    def setUp(self):
        self.config = AuraConfig.instance

    def test_config(self):
        print(self._testMethodName)

        # Regex for validating directory paths; including non-existing and prefixed "../"
        directory_regex = r"^.*\/(?:\.?[^./]+)$"

        # Check if config is available
        self.assertIsNotNone(AuraConfig.instance.config_file_path)

        # Check if "install_dir" is a valid directory (is evaluated at runtime)
        self.assertTrue(os.path.isdir(self.config.confuse_config["install_dir"].get()))

        # Reference to confuse config
        cfg = self.config.config

        # general
        self.assertRegex(cfg.general.socket_dir, directory_regex)
        self.assertRegex(cfg.general.cache_dir, directory_regex)

        # log
        self.assertTrue(os.path.isdir(cfg.log.directory))
        self.assertIn(cfg.log.level, ["debug", "info", "warning", "error", "critical"])

        # monitoring
        self.assertTrue(
            cfg.monitoring.heartbeat.host == ""
            or validators.domain(cfg.monitoring.heartbeat.host)
            or validators.ipv4(cfg.monitoring.heartbeat.host)
        )

        # api
        self.assertTrue(validators.url(cfg.api.steering.status))
        self.assertTrue(validators.url(cfg.api.steering.calendar))
        self.assertTrue(validators.url(cfg.api.tank.status))
        tank_playlist_url = cfg.api.tank.playlist.replace("${ID}", "1")
        self.assertTrue(validators.url(tank_playlist_url))
        self.assertTrue(validators.url(cfg.api.engine.status))
        self.assertTrue(validators.url(cfg.api.engine.store_playlog))
        self.assertTrue(validators.url(cfg.api.engine.store_clock))
        engine_health_url = cfg.api.engine.store_health.replace("${ENGINE_NUMBER}", "1")
        self.assertTrue(validators.url(engine_health_url))

        # scheduler
        self.assertRegex(cfg.scheduler.audio.source_folder, directory_regex)
        self.assertRegex(cfg.scheduler.audio.source_extension, r"\.[a-zA-Z0-9]+$")
        self.assertRegex(cfg.scheduler.audio.playlist_folder, directory_regex)
        self.assertIsInstance(cfg.scheduler.audio.engine_latency_offset, float)

        self.assertIsInstance(cfg.scheduler.fetching_frequency, int)
        self.assertIsInstance(cfg.scheduler.scheduling_window_start, int)
        self.assertIsInstance(cfg.scheduler.scheduling_window_end, int)
        self.assertIsInstance(cfg.scheduler.preload_offset, int)

        self.assertIsInstance(cfg.scheduler.input_stream.buffer, float)

        self.assertIsInstance(cfg.scheduler.fade_in_time, float)
        self.assertIsInstance(cfg.scheduler.fade_out_time, float)


if __name__ == "__main__":
    unittest.main()
