#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2017-now() - The Aura Engine Team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import pathlib
import unittest

from aura_engine.base.config import AuraConfig
from aura_engine.core.channels import ChannelName, ChannelType
from aura_engine.resources import (
    ResourceClass,
    ResourceMapping,
    ResourceType,
    ResourceUtil,
)
from aura_engine.scheduling.domain import (
    Episode,
    Playlist,
    PlaylistItem,
    Show,
    Timeslot,
)


class TestEngineResources(unittest.TestCase):
    """
    Testing the engine resource utilities.
    """

    config = None

    def setUp(self):
        self.config = AuraConfig.instance.config

    def test_to_abs_path(self):
        print(self._testMethodName)

        path = ResourceUtil.to_abs_path("abc")
        self.assertEqual("/", path[0])

    def test_audio_store_path(self):
        print(self._testMethodName)

        path = ResourceUtil.audio_store_path()
        self.assertEqual("/engine-core/audio/source", path.split("..")[1])

    def test_playlist_folder_path(self):
        print(self._testMethodName)

        path = ResourceUtil.playlist_folder_path()
        self.assertEqual("/engine-core/audio/playlist", path.split("..")[1])

    def test_playlist_folder_path(self):
        print(self._testMethodName)

        path = ResourceUtil.source_to_filepath("file://some/file")
        self.assertEqual("/engine-core/audio/source/some/file.flac", path.split("..")[1])

    def test_resource_mapping(self):
        print(self._testMethodName)

        rm = ResourceMapping()
        ct = rm.type_for_resource(ResourceType.FILE)
        self.assertEqual(ChannelType.QUEUE, ct)
        ct = rm.type_for_resource(ResourceType.STREAM_HTTP)
        self.assertEqual(ChannelType.HTTP, ct)
        ct = rm.type_for_resource(ResourceType.LINE)
        self.assertEqual(ChannelType.LIVE, ct)
        ct = rm.type_for_resource(ResourceType.M3U)
        self.assertEqual(ChannelType.QUEUE, ct)
        ct = rm.type_for_resource(ResourceType.POOL)
        self.assertEqual(ChannelType.QUEUE, ct)

    def test_live_channel_for_resource(self):
        print(self._testMethodName)

        rm = ResourceMapping()
        live_channel: ChannelName = rm.live_channel_for_resource("line://0")
        self.assertEqual("LIVE_0", live_channel.name)
        self.assertEqual("aura_engine_line_in_0", live_channel.value)
        live_channel: ChannelName = rm.live_channel_for_resource("line://1")
        self.assertEqual("LIVE_1", live_channel.name)
        self.assertEqual("aura_engine_line_in_1", live_channel.value)
        live_channel: ChannelName = rm.live_channel_for_resource("line://2")
        self.assertEqual("LIVE_2", live_channel.name)
        self.assertEqual("aura_engine_line_in_2", live_channel.value)
        live_channel: ChannelName = rm.live_channel_for_resource("line://3")
        self.assertEqual("LIVE_3", live_channel.name)
        self.assertEqual("aura_engine_line_in_3", live_channel.value)
        live_channel: ChannelName = rm.live_channel_for_resource("line://4")
        self.assertEqual("LIVE_4", live_channel.name)
        self.assertEqual("aura_engine_line_in_4", live_channel.value)

    def test_resource_class(self):
        print(self._testMethodName)

        rc_file = ResourceClass.FILE
        rc_live = ResourceClass.LIVE
        rc_stream = ResourceClass.STREAM
        rc_playlist = ResourceClass.PLAYLIST

        self.assertTrue(isinstance(rc_file.types, list))
        self.assertEqual(ResourceType.FILE, rc_file.types[0])
        self.assertEqual("fs", str(rc_file))
        self.assertEqual(0, rc_file.value.get("numeric"))
        self.assertEqual(rc_file.types, rc_file.value.get("types"))

        self.assertTrue(isinstance(rc_stream.types, list))
        self.assertEqual(ResourceType.STREAM_HTTP, rc_stream.types[0])
        self.assertEqual("http", str(rc_stream))
        self.assertEqual(1, rc_stream.value.get("numeric"))
        self.assertEqual(rc_stream.types, rc_stream.value.get("types"))

        self.assertTrue(isinstance(rc_live.types, list))
        self.assertEqual(ResourceType.LINE, rc_live.types[0])
        self.assertEqual("live", str(rc_live))
        self.assertEqual(2, rc_live.value.get("numeric"))
        self.assertEqual(rc_live.types, rc_live.value.get("types"))

        self.assertTrue(isinstance(rc_playlist.types, list))
        self.assertEqual(ResourceType.M3U, rc_playlist.types[0])
        self.assertEqual(ResourceType.POOL, rc_playlist.types[1])
        self.assertEqual("playlist", str(rc_playlist))
        self.assertEqual(3, rc_playlist.value.get("numeric"))
        self.assertEqual(rc_playlist.types, rc_playlist.value.get("types"))

    def test_resource_util_get_content_type(self):
        print(self._testMethodName)

        rt = ResourceUtil.get_content_type("file://magic.resource")
        self.assertEqual(ResourceType.FILE, rt)
        rt = ResourceUtil.get_content_type("http://magic.resource")
        self.assertEqual(ResourceType.STREAM_HTTP, rt)
        rt = ResourceUtil.get_content_type("https://magic.resource")
        self.assertEqual(ResourceType.STREAM_HTTP, rt)
        rt = ResourceUtil.get_content_type("line://magic.resource")
        self.assertEqual(ResourceType.LINE, rt)
        rt = ResourceUtil.get_content_type("m3u://magic.resource")
        self.assertEqual(ResourceType.M3U, rt)
        rt = ResourceUtil.get_content_type("pool://magic.resource")
        self.assertEqual(ResourceType.POOL, rt)

    def test_resource_util_get_content_class(self):
        print(self._testMethodName)

        rc = ResourceUtil.get_content_class(ResourceType.FILE)
        self.assertEqual(ResourceClass.FILE, rc)
        rc = ResourceUtil.get_content_class(ResourceType.STREAM_HTTP)
        self.assertEqual(ResourceClass.STREAM, rc)
        rc = ResourceUtil.get_content_class(ResourceType.LINE)
        self.assertEqual(ResourceClass.LIVE, rc)
        rc = ResourceUtil.get_content_class(ResourceType.M3U)
        self.assertEqual(ResourceClass.PLAYLIST, rc)
        rc = ResourceUtil.get_content_class(ResourceType.POOL)
        self.assertEqual(ResourceClass.PLAYLIST, rc)

    def test_resource_util_gen_m3u(self):
        print(self._testMethodName)

        pl = Playlist(1, "some description")
        e1 = PlaylistItem("file://file1", 2, 100, PlaylistItem.Metadata("a1", "b1", "c1"))
        e2 = PlaylistItem("file://file2", 3, 100, PlaylistItem.Metadata("a2", "b2", "c2"))
        e3 = PlaylistItem("file://file3", 5, 100, PlaylistItem.Metadata("a3", "b3", "c3"))
        pl.add(e1)
        pl.add(e2)
        pl.add(e3)

        file_path: str = "./.cache/test.m3u"
        ResourceUtil.generate_m3u_file(file_path, pl.items)
        print("created M3U file:" + file_path)

        self.assertTrue(pathlib.Path(file_path).is_file())

    def test_resource_util_get_items_string(self):
        print(self._testMethodName)

        ts = Timeslot(id=42, repetition_id=None, start=100000, end=100360, show=None, episode=None)
        pl = Playlist(1, "That's a real playlist with timeslot connection")
        e1 = PlaylistItem("file://file1", 2, 100, PlaylistItem.Metadata("a1", "b1", "c1"))
        e2 = PlaylistItem("file://file2", 3, 100, PlaylistItem.Metadata("a2", "b2", "c2"))
        e3 = PlaylistItem("file://file3", 5, 100, PlaylistItem.Metadata("a3", "b3", "c3"))
        pl.add(e1)
        pl.add(e2)
        pl.add(e3)
        ts.set_playlists(pl, None, None)

        items_str = ResourceUtil.get_items_string(pl.items)
        expect_str = "PlaylistItem [04:46:40 - 04:46:42 | 2.00sec | Source: ...file://file1], PlaylistItem [04:46:42 - 04:46:45 | 3.00sec | Source: ...file://file2], PlaylistItem [04:46:45 - 04:46:50 | 5.00sec | Source: ...file://file3]"
        print(items_str)
        self.assertEqual(expect_str, items_str)

    def test_generate_track_metadata(self):
        print(self._testMethodName)

        show = Show(id=123, name="some show")
        episode = Episode(id="1432", title="SuperSonicShow 23/1", memo="")
        ts = Timeslot(
            id=42, repetition_id=None, start=100000, end=100360, show=show, episode=episode
        )
        pl = Playlist(1, "That's a real playlist with timeslot connection")

        e1 = PlaylistItem(
            "file://file1", 5, 100, PlaylistItem.Metadata("artist", "album", "track1")
        )
        e2 = PlaylistItem(
            "file://file2", 5, 100, PlaylistItem.Metadata("artist", "album", "track2")
        )
        e3 = PlaylistItem(
            "file://file3", 5, 100, PlaylistItem.Metadata("artist", "album", "track3")
        )
        pl.add(e1)
        pl.add(e2)
        pl.add(e3)
        ts.set_playlists(pl, None, None)

        # Test playing track without track_start assignment
        e1.play.set_playing()
        metadata = ResourceUtil.generate_track_metadata(e1, False)
        self.assertEqual("", metadata.get("track_start"))

        # Test playing track without track_start assignment
        e1.play.set_playing()
        e1.play.play_start = 1710445690.0
        metadata = ResourceUtil.generate_track_metadata(e1, True)
        self.assertEqual("1970/01/02 04:46:40", metadata.get("track_start"))

        # Test other generated metadata
        metadata = ResourceUtil.generate_track_metadata(e3, False)
        self.assertEqual("some show", metadata.get("show_name"))
        self.assertEqual(123, metadata.get("show_id"))
        self.assertEqual(42, metadata.get("timeslot_id"))
        self.assertEqual(1, metadata.get("playlist_id"))
        self.assertEqual("3.0", metadata.get("playlist_item"))
        self.assertEqual(0, metadata.get("track_type"))
        self.assertEqual("artist", metadata.get("track_artist"))
        self.assertEqual("album", metadata.get("track_album"))
        self.assertEqual("track3", metadata.get("track_title"))


if __name__ == "__main__":
    unittest.main()
