#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2017-now() - The Aura Engine Team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import unittest

from aura_engine.base.config import AuraConfig
from aura_engine.base.logger import AuraLogger
from aura_engine.core.channels import GenericChannel
from aura_engine.resources import ResourceType
from aura_engine.scheduling.domain import (
    Episode,
    Playlist,
    PlaylistItem,
    PlaylistType,
    Show,
    Timeslot,
)


class TestDomain(unittest.TestCase):
    """
    Testing the Domain Models.
    """

    aura_logger = None

    def setUp(self):
        self.config = AuraConfig.instance.config
        self.aura_logger = AuraLogger(self.config)

    def test_timeslot_init(self):
        print(self._testMethodName)

        show = Show(id=555, name="3angulation")
        episode = Episode(id=3, title="SuperSonicShow 23/1", memo="")
        ts = Timeslot(
            id=333, repetition_id=888, start=1675463003, end=1675466003, show=show, episode=episode
        )

        ts.set_playlists(Playlist(1, "some description"), Playlist(2, ""), Playlist(3, ""))

    def test_timeslot_to_string(self):
        print(self._testMethodName)

        show = Show(id=555, name="3angulation")
        ts = Timeslot(id=333, repetition_id=888, start=1, end=2, show=show, episode=None)

        self.assertEqual("ID#333: 01:00:01 - 01:00:02 [Show#555: 3angulation]", str(ts))

    def test_playlist_init(self):
        print(self._testMethodName)

        pl = Playlist(1, "some description")

        pl.id = 333
        pl.timeslot = Timeslot(1, None, 123, 456, None, None)
        pl.type = PlaylistType.TIMESLOT

    def test_assign_playlists_to_timeslot(self):
        print(self._testMethodName)

        ts = Timeslot(
            id=333, repetition_id=888, start=1675463003, end=1675466009, show=None, episode=None
        )

        pl_timeslot = Playlist(1, "some description")
        pl_schedule = Playlist(2, "some description")
        pl_show = Playlist(3, "some description")

        ts.set_playlists(pl_timeslot, None, None)
        ts.set_playlists(None, pl_schedule, None)
        ts.set_playlists(None, None, pl_show)
        ts.set_playlists(pl_timeslot, pl_schedule, pl_show)
        self.assertIsNotNone(ts.playlists.timeslot)
        self.assertIsNotNone(ts.playlists.schedule)
        self.assertIsNotNone(ts.playlists.show)
        self.assertEqual(1675463003, ts.get_start())
        self.assertEqual(1675466009, ts.get_end())
        self.assertEqual(1675466009 - 1675463003, ts.get_duration())
        self.assertEqual(PlaylistType.TIMESLOT, ts.playlists.timeslot.get_type())
        self.assertEqual(PlaylistType.SCHEDULE, ts.playlists.schedule.get_type())
        self.assertEqual(PlaylistType.SHOW, ts.playlists.show.get_type())

    def test_get_current_playlist_from_timeslot(self):
        print(self._testMethodName)

        ts = Timeslot(
            id=333, repetition_id=888, start=1675463003, end=1675466009, show=None, episode=None
        )

        pl_timeslot = Playlist(1, "playlist A")
        pl_schedule = Playlist(2, "playlist B")
        pl_show = Playlist(3, "playlist C")

        ts.set_playlists(pl_timeslot, pl_schedule, pl_show)
        self.assertIsNotNone(ts.playlists.timeslot)
        self.assertIsNotNone(ts.playlists.schedule)
        self.assertIsNotNone(ts.playlists.show)
        self.assertEqual(pl_timeslot, ts.get_current_playlist())

        ts.set_playlists(pl_timeslot, None, None)
        self.assertIsNotNone(ts.playlists.timeslot)
        self.assertIsNone(ts.playlists.schedule)
        self.assertIsNone(ts.playlists.show)
        self.assertEqual(pl_timeslot, ts.get_current_playlist())

        ts.set_playlists(None, pl_schedule, pl_show)
        self.assertIsNone(ts.playlists.timeslot)
        self.assertIsNotNone(ts.playlists.schedule)
        self.assertIsNotNone(ts.playlists.show)
        self.assertEqual(pl_schedule, ts.get_current_playlist())

        ts.set_playlists(None, None, pl_show)
        self.assertIsNone(ts.playlists.timeslot)
        self.assertIsNone(ts.playlists.schedule)
        self.assertIsNotNone(ts.playlists.show)
        self.assertEqual(pl_show, ts.get_current_playlist())

    def test_get_current_playlist_from_virtual_timeslot(self):
        print(self._testMethodName)

        ts = Timeslot(
            id=333,
            virtual=True,
            repetition_id=888,
            start=1675463003,
            end=1675466009,
            show=None,
            episode=None,
        )

        # Virtual timeslots do use playlists
        self.assertTrue(ts.is_virtual())
        self.assertIsNone(ts.get_current_playlist())

        pl_timeslot = Playlist(1, "playlist A")
        pl_schedule = Playlist(2, "playlist B")
        pl_show = Playlist(3, "playlist C")

        ts.set_playlists(pl_timeslot, pl_schedule, pl_show)
        self.assertIsNotNone(ts.playlists.timeslot)
        self.assertIsNotNone(ts.playlists.schedule)
        self.assertIsNotNone(ts.playlists.show)

        # Even though there are playlists assigned, they are not returned
        self.assertIsNone(ts.get_current_playlist())

    def test_playlist_duration_zero(self):
        print(self._testMethodName)

        pl = Playlist(1, "some description")
        self.assertEqual(0, pl.get_duration())

    def test_playlist_add_items(self):
        print(self._testMethodName)

        pl = Playlist(1, "some description")
        e1 = PlaylistItem("file1.flac", 2, 100, PlaylistItem.Metadata("a1", "b1", "c1"))
        e2 = PlaylistItem("file2.flac", 3, 100, PlaylistItem.Metadata("a2", "b2", "c2"))
        e3 = PlaylistItem("file3.flac", 5, 100, PlaylistItem.Metadata("a3", "b3", "c3"))

        pl.add(e1)
        pl.add(e2)
        pl.add(e3)
        self.assertEqual(3, len(pl.get_items()))

    def test_playlist_remove_item(self):
        print(self._testMethodName)

        pl = Playlist(1, "some description")
        e1 = PlaylistItem("file1.flac", 2, 100, PlaylistItem.Metadata("a1", "b1", "c1"))
        e2 = PlaylistItem("file2.flac", 3, 100, PlaylistItem.Metadata("a2", "b2", "c2"))
        e3 = PlaylistItem("file3.flac", 5, 100, PlaylistItem.Metadata("a3", "b3", "c3"))

        pl.add(e1)
        pl.add(e2)
        pl.add(e3)
        self.assertEqual(3, len(pl.get_items()))

        e2.remove()
        self.assertEqual(2, len(pl.get_items()))
        self.assertEqual(e1, e3.prev)
        self.assertEqual(e3, e1.next)
        self.assertIsNone(e2.playlist)

        e3.remove()
        self.assertEqual(1, len(pl.get_items()))
        self.assertIsNone(e1.next)
        self.assertIsNone(e1.prev)
        self.assertIsNone(e3.next)
        self.assertIsNone(e3.prev)

    def test_playlist_duration(self):
        print(self._testMethodName)

        pl = Playlist(1, "some description")
        e1 = PlaylistItem("file1.flac", 2, 100, PlaylistItem.Metadata("a1", "b1", "c1"))
        e2 = PlaylistItem("file2.flac", 3, 100, PlaylistItem.Metadata("a2", "b2", "c2"))
        e3 = PlaylistItem("file3.flac", 5, 100, PlaylistItem.Metadata("a3", "b3", "c3"))

        pl.add(e1)
        pl.add(e2)
        pl.add(e3)
        self.assertEqual(10, pl.get_duration())

    def test_playlist_item_positions(self):
        print(self._testMethodName)

        pl = Playlist(1, "some description")
        e1 = PlaylistItem("file1.flac", 2, 100, PlaylistItem.Metadata("a1", "b1", "c1"))
        e2 = PlaylistItem("file2.flac", 3, 100, PlaylistItem.Metadata("a2", "b2", "c2"))
        e3 = PlaylistItem("file3.flac", 5, 100, PlaylistItem.Metadata("a3", "b3", "c3"))

        pl.add(e1)
        pl.add(e2)
        pl.add(e3)
        self.assertEqual(1, e1.get_position())
        self.assertEqual(2, e2.get_position())
        self.assertEqual(3, e3.get_position())

    def test_playlist_all_prev_and_next(self):
        print(self._testMethodName)

        ts = Timeslot(id=333, repetition_id=888, start=100, end=120, show=None, episode=None)

        pl = Playlist(1, "some description")
        e1 = PlaylistItem("file1.flac", 2, 100, PlaylistItem.Metadata("a1", "b1", "c1"))
        e2 = PlaylistItem("file2.flac", 3, 100, PlaylistItem.Metadata("a2", "b2", "c2"))
        e3 = PlaylistItem("file3.flac", 5, 100, PlaylistItem.Metadata("a3", "b3", "c3"))
        e4 = PlaylistItem("file4.flac", 7, 100, PlaylistItem.Metadata("a4", "b4", "c4"))
        e5 = PlaylistItem("file5.flac", 9, 100, PlaylistItem.Metadata("a5", "b5", "c5"))
        e6 = PlaylistItem("file6.flac", 11, 100, PlaylistItem.Metadata("a6", "b6", "c6"))
        ts.set_playlists(pl, None, None)

        pl.add(e1)
        pl.add(e2)
        pl.add(e3)
        pl.add(e4)
        pl.add(e5)
        pl.add(e6)
        prev = e3.get_all_prev()
        self.assertEqual(2, len(prev))
        next = e3.get_all_next(True)
        self.assertEqual(2, len(next))
        next = e3.get_all_next(False)
        self.assertEqual(3, len(next))
        self.assertEqual(e4.source, next[0].source)
        self.assertEqual(e5.source, next[1].source)
        self.assertEqual(e6.source, next[2].source)

    def test_playlist_item_start_end(self):
        print(self._testMethodName)

        ts = Timeslot(id=333, repetition_id=888, start=100, end=200, show=None, episode=None)

        pl = Playlist(1, "some description")
        e1 = PlaylistItem("file1.flac", 2, 100, PlaylistItem.Metadata("a1", "b1", "c1"))
        e2 = PlaylistItem("file2.flac", 3, 100, PlaylistItem.Metadata("a2", "b2", "c2"))
        e3 = PlaylistItem("file3.flac", 5, 100, PlaylistItem.Metadata("a3", "b3", "c3"))
        ts.set_playlists(pl, None, None)

        pl.add(e1)
        pl.add(e2)
        pl.add(e3)
        self.assertEqual(100, e1.get_start())
        self.assertEqual(102, e1.get_end())
        self.assertEqual(102, e2.get_start())
        self.assertEqual(105, e2.get_end())
        self.assertEqual(105, e3.get_start())
        self.assertEqual(110, e3.get_end())

    def test_playlist_item_content_type(self):
        print(self._testMethodName)

        e1 = PlaylistItem("file://file1.flac", 2, 100, PlaylistItem.Metadata("a1", "b1", "c1"))
        e2 = PlaylistItem(
            "https://stream.your.radio", 3, 100, PlaylistItem.Metadata("a2", "b2", "c2")
        )
        e3 = PlaylistItem("line://", 5, 100, PlaylistItem.Metadata("a3", "b3", "c3"))
        e4 = PlaylistItem("m3u://", 5, 100, PlaylistItem.Metadata("a3", "b3", "c3"))
        e5 = PlaylistItem("pool://", 5, 100, PlaylistItem.Metadata("a3", "b3", "c3"))
        e6 = PlaylistItem("unknown", 5, 100, PlaylistItem.Metadata("a3", "b3", "c3"))
        e7 = PlaylistItem(None, 5, 100, PlaylistItem.Metadata("a3", "b3", "c3"))

        self.assertEqual(ResourceType.FILE, e1.get_content_type())
        self.assertEqual(ResourceType.STREAM_HTTP, e2.get_content_type())
        self.assertEqual(ResourceType.LINE, e3.get_content_type())
        self.assertEqual(ResourceType.M3U, e4.get_content_type())
        self.assertEqual(ResourceType.POOL, e5.get_content_type())
        self.assertEqual(None, e6.get_content_type())
        self.assertEqual(None, e7.get_content_type())

    def test_playlist_item_to_string(self):
        print(self._testMethodName)

        ts = Timeslot(id=333, repetition_id=888, start=100, end=200, show=None, episode=None)
        pl = Playlist(1, "some description")
        item = PlaylistItem(
            "https://stream.your.radio", 3, 100, PlaylistItem.Metadata("a2", "b2", "c2")
        )

        ts.set_playlists(pl, None, None)
        pl.add(item)
        self.assertEqual(
            "PlaylistItem [01:01:40 - 01:01:43 | 3.00sec | Source: ...https://stream.your.radio]",
            str(item),
        )

    def test_playlist_item_play_state(self):
        print(self._testMethodName)

        item = PlaylistItem("file1.flac", 2, 100, PlaylistItem.Metadata("a1", "b1", "c1"))

        self.assertFalse(item.play.is_loading())
        self.assertFalse(item.play.is_ready())
        self.assertFalse(item.play.is_playing())
        self.assertFalse(item.play.is_done())
        self.assertIsNone(item.play.channel)

        item.play.set_loading(GenericChannel(1, "main", None))

        self.assertTrue(item.play.is_loading())
        self.assertFalse(item.play.is_ready())
        self.assertFalse(item.play.is_playing())
        self.assertFalse(item.play.is_done())
        self.assertIsNotNone(item.play.channel)

        item.play.set_ready()

        self.assertFalse(item.play.is_loading())
        self.assertTrue(item.play.is_ready())
        self.assertFalse(item.play.is_playing())
        self.assertFalse(item.play.is_done())
        self.assertIsNone(item.play.play_start)

        item.play.set_playing()

        self.assertFalse(item.play.is_loading())
        self.assertFalse(item.play.is_ready())
        self.assertTrue(item.play.is_playing())
        self.assertFalse(item.play.is_done())
        self.assertIsNotNone(item.play.play_start)
        item.play.play_start = 1710445690.0
        self.assertEqual(1710445690.0, item.play.get_play_start())

        item.play.set_done()

        self.assertFalse(item.play.is_loading())
        self.assertFalse(item.play.is_ready())
        self.assertFalse(item.play.is_playing())
        self.assertTrue(item.play.is_done())

    def test_update_playlist(self):
        print(self._testMethodName)

        ts = Timeslot(id=333, repetition_id=888, start=100, end=120, show=None, episode=None)

        # Create existing playlists
        pl_timeslot = Playlist(1, "some description")
        e1 = PlaylistItem("file1.flac", 2, 100, PlaylistItem.Metadata("a1", "b1", "c1"))
        e2 = PlaylistItem("file2.flac", 3, 100, PlaylistItem.Metadata("a2", "b2", "c2"))
        e3 = PlaylistItem("file3.flac", 5, 100, PlaylistItem.Metadata("a3", "b3", "c3"))
        e4 = PlaylistItem("file4.flac", 7, 100, PlaylistItem.Metadata("a4", "b4", "c4"))
        e5 = PlaylistItem("file5.flac", 9, 100, PlaylistItem.Metadata("a5", "b5", "c5"))
        e6 = PlaylistItem("file6.flac", 11, 100, PlaylistItem.Metadata("a6", "b6", "c6"))
        pl_timeslot.add(e1)
        pl_timeslot.add(e2)
        pl_timeslot.add(e3)
        pl_timeslot.add(e4)
        pl_timeslot.add(e5)
        pl_timeslot.add(e6)

        pl_show = Playlist(1, "some description")
        e1 = PlaylistItem("file1_show.flac", 2, 100, PlaylistItem.Metadata("a1s", "b1s", "c1s"))
        e2 = PlaylistItem("file2_show.flac", 3, 100, PlaylistItem.Metadata("a2s", "b2s", "c2s"))
        e3 = PlaylistItem("file3_show.flac", 5, 100, PlaylistItem.Metadata("a3s", "b3s", "c3s"))
        pl_show.add(e1)
        pl_show.add(e2)
        pl_show.add(e3)

        ts.set_playlists(pl_timeslot, None, pl_show)

        # Create new remote playlist
        pl_timeslot_new = Playlist(1, "some description")
        e1 = PlaylistItem(
            "file://file1_new.flac", 2, 100, PlaylistItem.Metadata("a1_new", "b1", "c1")
        )
        e2 = PlaylistItem(
            "https://stream.your.radio", 3, 100, PlaylistItem.Metadata("a2_new", "b2", "c2")
        )
        e3 = PlaylistItem("line://", 5, 100, PlaylistItem.Metadata("a3_new", "b3", "c3"))
        e4 = PlaylistItem("m3u://", 5, 100, PlaylistItem.Metadata("a4_new", "b3", "c3"))
        pl_timeslot_new.add(e1)
        pl_timeslot_new.add(e2)
        pl_timeslot_new.add(e3)
        pl_timeslot_new.add(e4)

        # Test initial state
        self.assertEqual(6, len(pl_timeslot.items))
        self.assertEqual("a1", pl_timeslot.items[0].metadata.artist)
        self.assertEqual("file3.flac", pl_timeslot.items[2].source)

        # Test playlist update
        pl_timeslot.update_playlist(pl_timeslot_new)
        self.assertEqual(4, len(pl_timeslot.items))
        self.assertEqual("a1_new", pl_timeslot.items[0].metadata.artist)
        self.assertEqual("line://", pl_timeslot.items[2].source)

    def test_update_timeslot(self):
        print(self._testMethodName)

        show1 = Show(1, "show1")
        show2 = Show(2, "show2")
        episode1 = Episode(11, "title1", "my memo for this episode")
        episode2 = Episode(22, "title2", None)
        ts1 = Timeslot(id=333, repetition_id=888, start=100, end=120, show=show1, episode=episode1)
        ts2 = Timeslot(id=444, repetition_id=999, start=100, end=120, show=show2, episode=episode2)

        # Create existing playlists
        pl_timeslot = Playlist(1, "some description")
        e1 = PlaylistItem("file1.flac", 2, 100, PlaylistItem.Metadata("a1", "b1", "c1"))
        e2 = PlaylistItem("file2.flac", 3, 100, PlaylistItem.Metadata("a2", "b2", "c2"))
        e3 = PlaylistItem("file3.flac", 5, 100, PlaylistItem.Metadata("a3", "b3", "c3"))
        e4 = PlaylistItem("file4.flac", 7, 100, PlaylistItem.Metadata("a4", "b4", "c4"))
        e5 = PlaylistItem("file5.flac", 9, 100, PlaylistItem.Metadata("a5", "b5", "c5"))
        e6 = PlaylistItem("file6.flac", 11, 100, PlaylistItem.Metadata("a6", "b6", "c6"))
        pl_timeslot.add(e1)
        pl_timeslot.add(e2)
        pl_timeslot.add(e3)
        pl_timeslot.add(e4)
        pl_timeslot.add(e5)
        pl_timeslot.add(e6)

        pl_show = Playlist(2, "some description")
        e1 = PlaylistItem("file1_show.flac", 2, 100, PlaylistItem.Metadata("a1s", "b1s", "c1s"))
        e2 = PlaylistItem("file2_show.flac", 3, 100, PlaylistItem.Metadata("a2s", "b2s", "c2s"))
        e3 = PlaylistItem("file3_show.flac", 5, 100, PlaylistItem.Metadata("a3s", "b3s", "c3s"))
        pl_show.add(e1)
        pl_show.add(e2)
        pl_show.add(e3)

        pl_timeslot2 = Playlist(1, "some description")
        e1 = PlaylistItem("file1_fresh.flac", 2, 100, PlaylistItem.Metadata("a1", "b1", "c1"))
        e2 = PlaylistItem("file2_fresh.flac", 3, 100, PlaylistItem.Metadata("a2", "b2", "c2"))
        e3 = PlaylistItem("file3_fresh.flac", 5, 100, PlaylistItem.Metadata("a3", "b3", "c3"))
        e4 = PlaylistItem("file4_fresh.flac", 7, 100, PlaylistItem.Metadata("a4", "b4", "c4"))
        e5 = PlaylistItem("file5_fresh.flac", 9, 100, PlaylistItem.Metadata("a5", "b5", "c5"))
        e6 = PlaylistItem("file6_fresh.flac", 11, 100, PlaylistItem.Metadata("a6", "b6", "c6"))
        e7 = PlaylistItem("file7_fresh.flac", 11, 100, PlaylistItem.Metadata("a7", "b7", "c7"))
        pl_timeslot2.add(e1)
        pl_timeslot2.add(e2)
        pl_timeslot2.add(e3)
        pl_timeslot2.add(e4)
        pl_timeslot2.add(e5)
        pl_timeslot2.add(e6)
        pl_timeslot2.add(e7)

        pl_schedule = Playlist(3, "some fresh description")
        e1 = PlaylistItem(
            "fresh_schedule_file1.flac",
            2,
            100,
            PlaylistItem.Metadata("a1fresh", "b1fresh", "c1fresh"),
        )
        e2 = PlaylistItem(
            "fresh_schedule_file2.flac",
            3,
            100,
            PlaylistItem.Metadata("a2fresh", "b2fresh", "c2fresh"),
        )
        e3 = PlaylistItem(
            "fresh_schedule_file3.flac",
            5,
            100,
            PlaylistItem.Metadata("a3fresh", "b3fresh", "c3fresh"),
        )
        e4 = PlaylistItem(
            "fresh_schedule_file4.flac",
            15,
            100,
            PlaylistItem.Metadata("a4fresh", "b4fresh", "c4fresh"),
        )
        pl_schedule.add(e1)
        pl_schedule.add(e2)
        pl_schedule.add(e3)
        pl_schedule.add(e4)

        ts1.set_playlists(pl_timeslot, None, pl_show)
        ts2.set_playlists(pl_timeslot2, pl_schedule, None)

        # Test initial state
        self.assertEqual(333, ts1.get_id())
        self.assertEqual(888, ts1.get_repetition_id())
        self.assertEqual(1, ts1.get_show().id)
        self.assertEqual("title1", ts1.get_episode().title)
        self.assertEqual(6, len(ts1.playlists.get(PlaylistType.TIMESLOT).items))
        self.assertIsNone(ts1.playlists.get(PlaylistType.SCHEDULE))
        self.assertEqual(3, len(ts1.playlists.get(PlaylistType.SHOW).items))
        self.assertEqual(6, len(ts1.get_current_playlist().items))

        # Do update
        ts1.update(ts2)

        # Test timeslot update
        self.assertEqual(444, ts1.get_id())
        self.assertEqual(999, ts1.get_repetition_id())
        self.assertEqual(2, ts1.get_show().id)
        self.assertEqual("title2", ts1.get_episode().title)
        self.assertEqual(7, len(ts1.playlists.get(PlaylistType.TIMESLOT).items))
        self.assertEqual(4, len(ts1.playlists.get(PlaylistType.SCHEDULE).items))
        self.assertIsNone(ts1.playlists.get(PlaylistType.SHOW))
        self.assertEqual(7, len(ts1.get_current_playlist().items))

        # Do update without a timeslot playlist
        ts2.playlists.set(PlaylistType.TIMESLOT, None)
        ts1.update(ts2)

        # Test timeslot update
        self.assertEqual(444, ts1.get_id())
        self.assertEqual(999, ts1.get_repetition_id())
        self.assertEqual(2, ts1.get_show().id)
        self.assertEqual("title2", ts1.get_episode().title)
        self.assertIsNone(ts1.playlists.get(PlaylistType.TIMESLOT))
        self.assertEqual(4, len(ts1.playlists.get(PlaylistType.SCHEDULE).items))
        self.assertIsNone(ts1.playlists.get(PlaylistType.SHOW))
        self.assertEqual(4, len(ts1.get_current_playlist().items))


if __name__ == "__main__":
    unittest.main()
