#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2017-now() - The Aura Engine Team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import inspect
import time
import unittest

from aura_engine.base.config import AuraConfig
from aura_engine.base.utils import SimpleUtil as SU
from aura_engine.control import EngineExecutor
from aura_engine.core.channels import GenericChannel
from aura_engine.engine import Engine
from aura_engine.scheduling.domain import Playlist, PlaylistItem, Timeslot
from aura_engine.scheduling.scheduler import AuraScheduler
from aura_engine.scheduling.utils import TimetableRenderer


class CommandRegistry:
    commands: {} = {}

    def register(f, cmd: EngineExecutor):
        if not CommandRegistry.commands.get(f):
            CommandRegistry.commands[f] = []
        CommandRegistry.commands.get(f).append(cmd)

    def get(f):
        return CommandRegistry.commands.get(f)

    def reset(f):
        c: EngineExecutor
        if CommandRegistry.commands.get(f):
            for c in CommandRegistry.commands.get(f):
                c.cancel()
        CommandRegistry.commands[f] = None

    def prune():
        for key in CommandRegistry.commands.keys():
            CommandRegistry.reset(key)


#
# Mocks
#


class MockedPlayer:
    """
    Mocked version of engine.
    """

    def roll(channel, seconds_to_roll: int):
        print(f"called pre-rolling for {channel} seconds on player channel {seconds_to_roll}")
        pass


class MockedEngine:
    """
    Mocked version of engine.
    """

    scheduler: AuraScheduler
    player: MockedPlayer

    def __init__(self):
        """
        Init.
        """
        self.scheduler = None
        self.player = MockedPlayer


class MockedTimeslotCommand:
    """
    Command for triggering start and end of timeslot events.
    """

    engine: Engine = None
    config: AuraConfig = None
    timeslot: Timeslot

    def __init__(self, engine: Engine, timeslot: Timeslot):
        """
        Initialize the timeslot command.

        Args:
            engine (Engine): The engine
            timeslot (Timeslot): The timeslot which is starting at this time
        """
        self.config = AuraConfig.instance.config
        self.engine = engine
        caller = inspect.stack()[2].function
        CommandRegistry.register(caller, self)
        print(f"Created instance of MockedTimeslotCommand from {caller}")
        self.timeslot = timeslot
        self.do_start_timeslot(timeslot)
        self.do_end_timeslot(timeslot)

    def do_start_timeslot(self, timeslot: Timeslot):
        """
        Indicate the start of the timeslot by sending a `on_timeslot_start` event.
        """
        caller = inspect.stack()[2].function
        print(f"Called do_play of MockedTimeslotCommand from {caller}")

    def do_end_timeslot(self, timeslot: Timeslot):
        """
        Indicate the start of the timeslot by sending a `on_timeslot_end` event.

        Also resetting the used channel.
        """
        caller = inspect.stack()[2].function
        print(f"Called do_play of MockedTimeslotCommand from {caller}")

    def cancel(self):
        """
        Cancel the command.
        """
        pass


class MockedPlayCommand:
    """
    Mocked command for triggering timed preloading and playing.
    """

    engine: Engine = None
    config: AuraConfig = None
    items: list[PlaylistItem]

    def __init__(self, engine: Engine, items: list[PlaylistItem]):
        """
        Initialize the play command.

        Args:
            engine (Engine): The engine
            items ([PlaylistItem]): One or more playlist items to be started
        """
        self.config = AuraConfig.instance.config
        self.engine = engine
        caller = inspect.stack()[2].function
        CommandRegistry.register(caller, self)
        print(f"Created instance of MockedPlayCommand from {caller}")
        self.items = items
        self.do_preload(items)
        self.do_play(items)

    def do_preload(self, items: list[PlaylistItem]):
        """
        Preload the items.

        Args:
            items ([PlaylistItem]): The set of playlist items to be pre-loaded.
        """
        caller = inspect.stack()[2].function
        print(f"Called do_preload of MockedPlayCommand from {caller}")
        items[0].play_channel = GenericChannel(1, "main channel", None)

    def do_play(self, items: list[PlaylistItem]):
        """
        Play the items.

        Args:
            items ([PlaylistItem]): The set of playlist items to be played.
        """
        caller = inspect.stack()[2].function
        print(f"Called do_play of MockedPlayCommand from {caller}")

    def cancel(self):
        """
        Cancel the command.
        """
        pass


class TestSchedulingScheduler(unittest.TestCase):
    """
    Testing the scheduling utils.
    """

    engine: Engine
    config: AuraConfig

    # Setup and teardown

    def setUp(self):
        self.config = AuraConfig.instance.config
        self.engine = Engine.get_instance()

        # Init mocked command classes
        AuraScheduler.TimeslotCommandClass = MockedTimeslotCommand
        AuraScheduler.PlayCommandClass = MockedPlayCommand

    def tearDown(self):
        pass

    #
    # Utilities
    #

    def build_timetable(
        self, scheduler: AuraScheduler, current_start, current_end, st="file://"
    ) -> list[Timeslot]:
        # Delete the timetable file to isolate test case
        f = scheduler.timetable.timetable_file
        scheduler.timetable.timetable_file = f.replace("timetable.json", "timetable-test.json")
        #

        # Prune command registry
        CommandRegistry.prune()

        # Build some timetable
        now = SU.timestamp()
        ts1 = Timeslot(
            id=1,
            repetition_id=None,
            start=now - 500,
            end=now + current_start,
            show=None,
            episode=None,
        )
        ts2 = Timeslot(
            id=2,
            repetition_id=None,
            start=now + current_start,
            end=now + current_end,
            show=None,
            episode=None,
        )
        ts3 = Timeslot(
            id=3,
            repetition_id=None,
            start=now + current_end,
            end=now + 540,
            show=None,
            episode=None,
        )
        ts4 = Timeslot(
            id=4, repetition_id=None, start=now + 540, end=now + 720, show=None, episode=None
        )

        # Playlist for timeslot 2
        pl2 = Playlist("9999", "Playlist XYZ", False)
        alpha = PlaylistItem(st + "alpha.flac", 100, 100, None)
        beta = PlaylistItem(st + "beta.flac", current_end - current_start - 200, 100, None)
        gamma = PlaylistItem(st + "gamma.flac", 50, 100, None)
        delta = PlaylistItem(st + "delta.flac", 50, 100, None)
        pl2.add(alpha)
        pl2.add(beta)
        pl2.add(gamma)
        pl2.add(delta)

        # Playlist for timeslot 3
        pl3 = Playlist("7777", "Colours", False)
        magenta = PlaylistItem(st + "magenta.flac", 100, 100, None)
        black = PlaylistItem(st + "black.flac", current_end - current_start - 200, 100, None)
        taupe = PlaylistItem(st + "taupe.flac", 100, 100, None)
        pl3.add(magenta)
        pl3.add(black)
        pl3.add(taupe)

        ts2.set_playlists(None, pl2, None)
        ts3.set_playlists(pl3, None, None)
        return [ts1, ts2, ts3, ts4]

    #
    # Test Cases
    #

    def test_start_and_terminate_scheduler(self):
        print(self._testMethodName)

        # Boot the scheduler
        scheduler = AuraScheduler(self.engine)

        scheduler.boot()
        time.sleep(2)
        self.assertTrue(scheduler.is_alive())

        # Shut the scheduler down
        scheduler.terminate()
        time.sleep(10)
        self.assertFalse(scheduler.is_alive())

    def test_get_timetable(self):
        print(self._testMethodName)

        # Construct the scheduler
        scheduler = AuraScheduler(MockedEngine())

        # Build some timetable
        scheduler.timetable.timetable = self.build_timetable(scheduler, -400, +400)

        tt = scheduler.get_timetable()
        self.assertIsNotNone(tt)
        self.assertEqual(4, len(tt.timetable))

    def test_play_active_item_with_roll(self):
        print(self._testMethodName)

        # Construct the scheduler
        scheduler = AuraScheduler(MockedEngine())

        # Build some timetable
        scheduler.timetable.timetable = self.build_timetable(scheduler, -400, +400)

        # Display timetable for debugging
        tr = TimetableRenderer(scheduler)
        report = tr.get_ascii_timeslots()
        print(report)

        # Before we proceed, ensure there is really an active item
        active_item = scheduler.timetable.get_current_item()
        self.assertIsNotNone(active_item)

        scheduler.play_active_item()
        cmds = CommandRegistry.get(self._testMethodName)

        # Expecting at least one timeslot and a play command
        self.assertEqual(2, len(cmds))
        self.assertTrue(isinstance(cmds[0], MockedTimeslotCommand))
        self.assertTrue(isinstance(cmds[1], MockedPlayCommand))

        timeslot_cmd: MockedTimeslotCommand = cmds[0]
        play_cmd: MockedPlayCommand = cmds[1]
        self.assertEqual(2, timeslot_cmd.timeslot.get_id())
        self.assertEqual(1, len(play_cmd.items))
        channel: GenericChannel = play_cmd.items[0].play_channel
        self.assertEqual("main channel", channel.name)

    def test_play_active_item_no_roll(self):
        print(self._testMethodName)

        # Construct the scheduler
        scheduler = AuraScheduler(MockedEngine())

        # Build some timetable
        scheduler.timetable.timetable = self.build_timetable(scheduler, -400, +5)

        scheduler.play_active_item()
        cmds = CommandRegistry.get(self._testMethodName)

        # Expecting at least one timeslot command
        self.assertEqual(1, len(cmds))
        self.assertTrue(isinstance(cmds[0], MockedTimeslotCommand))

    def test_play_active_invalid_source(self):
        print(self._testMethodName)

        # Construct the scheduler
        scheduler = AuraScheduler(MockedEngine())

        # Build some timetable
        scheduler.timetable.timetable = self.build_timetable(scheduler, -400, +5, "3d_video://")

        scheduler.play_active_item()
        cmds = CommandRegistry.get(self._testMethodName)

        # Expecting at least one timeslot command
        self.assertEqual(1, len(cmds))
        self.assertTrue(isinstance(cmds[0], MockedTimeslotCommand))

    def test_queue_program_startup_items(self):
        print(self._testMethodName)

        # Construct the scheduler
        scheduler = AuraScheduler(MockedEngine())

        # Build some timetable
        scheduler.timetable.timetable = self.build_timetable(scheduler, -400, +400)

        scheduler.queue_startup_items()
        cmds = CommandRegistry.get("queue_startup_items")

        # Expecting only one play command, is the timeslot command was created
        # when scheduling the active item already
        self.assertEqual(1, len(cmds))
        self.assertTrue(isinstance(cmds[0], MockedPlayCommand))

        play_cmd: MockedPlayCommand = cmds[0]
        self.assertEqual(2, len(play_cmd.items))
        channel: GenericChannel = play_cmd.items[0].play_channel
        self.assertEqual("main channel", channel.name)

    def test_queue_program_nothing_next_in_window(self):
        print(self._testMethodName)

        # Construct the scheduler
        scheduler = AuraScheduler(MockedEngine())

        # Build some timetable
        scheduler.timetable.timetable = self.build_timetable(scheduler, -400, +400)

        scheduler.queue_program()
        cmds = CommandRegistry.get(self._testMethodName)

        # Expecting no command as everything is out of the scheduling window
        self.assertIsNone(cmds)

    def test_queue_program_valid_next_in_window(self):
        print(self._testMethodName)

        # Construct the scheduler
        scheduler = AuraScheduler(MockedEngine())

        # Build some timetable
        scheduler.timetable.timetable = self.build_timetable(scheduler, -400, +30)

        scheduler.queue_program()
        cmds = CommandRegistry.get(self._testMethodName)
        cmds += CommandRegistry.get("queue_program")

        # Expecting one timeslot and one play command for the next timeslot
        self.assertEqual(2, len(cmds))
        self.assertTrue(isinstance(cmds[0], MockedTimeslotCommand))
        self.assertTrue(isinstance(cmds[1], MockedPlayCommand))

        timeslot_cmd: MockedTimeslotCommand = cmds[0]
        play_cmd: MockedPlayCommand = cmds[1]
        self.assertEqual(3, timeslot_cmd.timeslot.get_id())
        self.assertEqual(3, len(play_cmd.items))
        channel: GenericChannel = play_cmd.items[0].play_channel
        self.assertEqual("main channel", channel.name)


if __name__ == "__main__":
    unittest.main()
