#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2017-now() - The Aura Engine Team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import unittest

from aura_engine.base.config import AuraConfig
from aura_engine.base.utils import SimpleUtil as SU
from aura_engine.scheduling.domain import Playlist, PlaylistItem, Timeslot
from aura_engine.scheduling.timetable import TimetableService
from aura_engine.scheduling.utils import M3UPlaylistProcessor, TimetableRenderer


class MockedScheduler:
    timetable: TimetableService

    def __init__(self, timetable: list[Timeslot]):
        self.timetable = timetable

    def get_timetable(self) -> TimetableService:
        return self.timetable


class TestSchedulingUtils(unittest.TestCase):
    """
    Testing the scheduling utils.
    """

    config = None
    m3u = None
    timetable: TimetableService

    # Setup and teardown

    def setUp(self):
        self.config = AuraConfig.instance.config
        self.m3u = M3UPlaylistProcessor()
        self.m3u.playlist_folder = "./tests/m3u/"

        cache_location = self.config.general.cache_dir
        self.timetable = TimetableService(cache_location)

    def tearDown(self):
        pass

    #
    # Test Cases
    #

    def test_spread_m3u_playlist(self):
        print(self._testMethodName)

        item = PlaylistItem("m3u://sample.m3u", None, 100, PlaylistItem.Metadata("", "", ""))
        items = self.m3u.spread(item)
        self.assertIsNotNone(items)
        self.assertEqual(4, len(items))
        self.assertEqual("file://music-lib/Unser Album/Dein Lied.ogg", items[1].source)
        self.assertEqual("Alle", items[1].metadata.artist)
        self.assertEqual("Dein Lied", items[1].metadata.title)

    def test_spread_no_m3u_item(self):
        print(self._testMethodName)

        item = PlaylistItem("file://no-m3u.mp3", None, 100, PlaylistItem.Metadata("", "", ""))
        items = self.m3u.spread(item)
        self.assertIsNotNone(items)
        self.assertEqual(1, len(items))
        self.assertEqual("file://no-m3u.mp3", items[0].source)

    def test_spread_no_m3u_playlist(self):
        print(self._testMethodName)

        item = PlaylistItem("m3u://404.m3u", None, 100, PlaylistItem.Metadata("", "", ""))
        items = self.m3u.spread(item)

        self.assertEqual(0, len(items))

    def test_timetable_renderer(self):
        print(self._testMethodName)

        # Build timeslots
        now = SU.timestamp()
        ts1 = Timeslot(
            id=1, repetition_id=None, start=now - 180, end=now - 120, show=None, episode=None
        )
        ts2 = Timeslot(
            id=1, repetition_id=None, start=now - 120, end=now + 120, show=None, episode=None
        )
        ts3 = Timeslot(
            id=1, repetition_id=None, start=now + 120, end=now + 180, show=None, episode=None
        )
        ts4 = Timeslot(
            id=1, repetition_id=None, start=now + 180, end=now + 360, show=None, episode=None
        )

        # Set playlist for timeslot 2 (currently playing)
        pl = Playlist("999", "Playlist X", False)
        alpha = PlaylistItem("alpha.flac", 20, 100, None)
        beta = PlaylistItem("beta.flac", 100, 100, None)
        gamma = PlaylistItem("gamma.flac", 20, 100, None)
        delta = PlaylistItem("delta.flac", 200, 100, None)
        epsilon = PlaylistItem("epsilon.flac", 300, 100, None)

        pl.add(alpha)
        pl.add(beta)
        pl.add(gamma)
        pl.add(delta)
        pl.add(epsilon)
        ts2.set_playlists(None, pl, None)

        # Set playlist for timeslot 3 (playing next)
        pl = Playlist("555", "Playlist Y", False)
        alpha = PlaylistItem("red.flac", 20, 100, None)
        beta = PlaylistItem("green.flac", 100, 100, None)
        gamma = PlaylistItem("blue.flac", 20, 100, None)
        pl.add(alpha)
        pl.add(beta)
        pl.add(gamma)
        ts3.set_playlists(pl, None, None)

        timeslots = [ts1, ts2, ts3, ts4]
        self.timetable.timetable = timeslots
        scheduler = MockedScheduler(self.timetable)

        # Build and display report
        tr = TimetableRenderer(scheduler)
        report = tr.get_ascii_timeslots()
        print(report)


if __name__ == "__main__":
    unittest.main()
